#include "arguments.h"

#include <iostream>
#include <utility>
#include <fstream>
#include <string>
#include <sstream>

#include "constants.h"
#include "input/file_input.h"
#include "output/file_output.h"
#include "input/console_input.h"
#include "output/console_output.h"
#include "input/curses_input.h"
#include "output/curses_output.h"
#include "rng.h"

feature proc_args(int argc, char **argv,
                  std::unique_ptr<cursor>  &curse,
                  std::unique_ptr<input>   &in,
                  std::unique_ptr<output> &out,
                  std::unique_ptr<RNG>     &rng) {
        feature result = 0;
        std::string str;
        std::string fn_fin;
        std::string fn_fout;
        std::string fn_lout;
        std::string fn_save;
        std::string seed;

        for (int i = 1; i < argc; ++i) {
                str = argv[i];

                if (str == "-n") {
                        if (result & (FEATURE_IN_FILE | FEATURE_OUT_FILE))
                                return FEATURE_CONFLICT | i;

                        result |= FEATURE_NCURSES;
                } else if (str == "-r") {
                        result |= FEATURE_RAND_MAP;
                } else if (str == "-c") {
                        result |= FEATURE_ENEMIES_CHASE;
                } else if (str == "-d") {
                        result |= FEATURE_DOORS;
                } else if (str == "-i") {
                        result |= FEATURE_INVENTORY;
                } else if (str == "-t") {
                        result |= FEATURE_THROW;
                } else if (str == "-R") {
                        result |= FEATURE_REVISIT;
                } else if (str == "-e") {
                        result |= FEATURE_EXTRA_STUFF;
                } else if (str == "-E") {
                        result |= FEATURE_EXTRA_LEVELS;
                } else if (str == "-o") {
                        result |= FEATURE_WALK_OVER;
                } else if (str == "-s") {
                        ++i;
                        str = argv[i];

                        if (!seed.empty() && seed != str)
                                return FEATURE_CONFLICT | i;

                        seed = str;
                        result |= FEATURE_SEED;
                } else if (str == "-I") {
                        ++i;
                        str = argv[i];

                        if (!fn_fin.empty() && fn_fin != str)
                                return FEATURE_CONFLICT | i;

                        fn_fin = str;
                        result |= FEATURE_IN_FILE;
                } else if (str == "-O") {
                        ++i;
                        str = argv[i];

                        if (!fn_fout.empty() && fn_fout != str)
                                return FEATURE_CONFLICT | i;

                        fn_fout = str;
                        result |= FEATURE_OUT_FILE;
                } else if (str == "-h" || str == "--help") {
                        return FEATURE_LIST_ARGS;
                } else {
                        return FEATURE_PANIC | i;
                }
        }

        if (result & FEATURE_IN_FILE) {
                std::ifstream fin(fn_fin);

                if (!fin.good())
                        return FEATURE_PANIC_FILE | FEATURE_IN_FILE;

                in = std::make_unique<file_input>(std::move(fin));
        } else if (!(result & FEATURE_NCURSES))
                in = std::make_unique<console_input>(std::cin);

        if (result & FEATURE_OUT_FILE) {
                std::ofstream fout(fn_fout);

                if (!fout.good())
                        return FEATURE_PANIC_FILE | FEATURE_OUT_FILE;

                out = std::make_unique<file_output>(std::move(fout));
        } else if (!(result & FEATURE_NCURSES))
                out = std::make_unique<console_output>(std::cout);

        if (result & FEATURE_NCURSES) {
                curse = std::make_unique<cursor>();
                in = std::make_unique<curses_input>(curse.get());
                out = std::make_unique<curses_output>(curse.get());
        }

        if (result & FEATURE_SEED) {
                std::istringstream iss {seed};
                unsigned int tmp;
                iss >> tmp;

                if (!iss.good())
                        return FEATURE_PANIC_SEED;

                rng = std::make_unique<RNG>(tmp);
        } else {
                rng = std::make_unique<RNG>();
        }

        return result;
}

// IMPORTANT: This is meant for already panicking
void panic_args(feature panic) {
        using namespace std;

        if (panic & FEATURE_PANIC)
                cerr << "Invalid argument No. " << (panic ^ FEATURE_PANIC)
                     << " !" << endl;
        else if (panic & FEATURE_CONFLICT)
                cerr << "Argument No. " << (panic ^ FEATURE_CONFLICT)
                     << " conflicts with a previous argument!" << endl;
        else if (panic & FEATURE_PANIC_FILE) {
                switch (panic ^ FEATURE_PANIC_FILE) {
                case FEATURE_IN_FILE:
                        cerr << "Cannot open specified input file!" << endl;
                        break;

                case FEATURE_OUT_FILE:
                        cerr << "Cannot open specified output file!" << endl;
                        break;

                default:
                        cerr << "Something must have went really, really, wrong..."
                             << endl;
                        break;
                }
        } else if (panic & FEATURE_PANIC_SEED) {
                cerr << "Invalid seed" << endl;
        } else
                cerr << "Something must have went really, really, wrong..."
                     << endl;
}

void print_args_list() {
        static const char *ARGS_LIST = "-n        : Use ncurses for I/O\n\
-r        : Randomly generate maps\n\
-c        : Enemies chase the player (CAUTION: THEY CAN REALLY CHASE!!!)\n\
-d        : Enemies can go through doors (CAUTION: DO NOT ENABLE WITH CHASING!)\n\
-i        : Enable inventory (player can pick up potions)\n\
-t        : Enable throw\n\
-R        : Enable revisiting levels\n\
-e        : Enable extra potions and races\n\
-E        : Enable extra levels\n\
-o        : Allows characters to go over gold and potions\n\
-s [seed] : Sets initial seed to seed\n\
-I [file] : Reads commands from file. CANNOT BE USED WITH -n.\n\
-O [file] : Outputs to file. CANNOT BE USED WITH -n.\n\
-h/--help : Displays options list (doesn't start a game)";
        std::cout << ARGS_LIST << std::endl;
}
